"""DAQ Module."""

import logging
import typing as t
from collections import namedtuple

import numpy as np
from zhinst.core import DataAcquisitionModule as ZIDAQModule

from zhinst.toolkit.driver.modules.base_module import BaseModule
from zhinst.toolkit.nodetree.helper import NodeDict

if t.TYPE_CHECKING:  # pragma: no cover
    from zhinst.toolkit.session import Session

logger = logging.getLogger(__name__)
DAQResult = namedtuple("DAQResult", ["header", "value", "time", "frequency", "shape"])


class DAQModule(BaseModule):
    """Data Acquisition Module.

    The Data Acquisition Module corresponds to the Data Acquisition tab of the
    LabOne User Interface. It enables the user to record and align time and
    frequency domain data from multiple instrument signal sources at a defined
    data rate. The data may be recorded either continuously or in bursts based
    upon trigger criteria analogous to the functionality provided by laboratory
    oscilloscopes.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/data_acquisition_module.html

    Args:
        daq_module: Instance of the core DAQ module.
        session: Session to the Data Server.
    """

    def __init__(self, daq_module: ZIDAQModule, session: "Session"):
        super().__init__(daq_module, session)
        self.root.update_nodes(
            {
                "/triggernode": {
                    "GetParser": self._get_node,
                    "SetParser": self._set_node,
                }
            },
            raise_for_invalid_node=False,
        )

    @staticmethod
    def _process_burst(
        node: str, burst: t.Dict[str, t.Any], clk_rate: float
    ) -> DAQResult:
        """Process a single burst into a formatted DAQResult object.

        Args:
            node: Name of the node of the burst.
            burst: raw burst data.
            clk_rate: Clock rate [Hz] for converting the timestamps. Only
                applies if the raw flag is reset.

        Returns:
                Processed and formatted burst data.
        """
        if "fft" in node:
            bin_count = len(burst["value"][0])
            bin_resolution = burst["header"]["gridcoldelta"]
            frequency = np.arange(bin_count)
            bandwidth = bin_resolution * len(frequency)
            frequency = frequency * bin_resolution
            if "xiy" in node:
                frequency = frequency - bandwidth / 2.0 + bin_resolution / 2.0
            return DAQResult(
                burst.get("header", {}),
                burst["value"],
                None,
                frequency,
                burst["value"].shape,
            )
        timestamp = burst["timestamp"]
        return DAQResult(
            burst.get("header", {}),
            burst["value"],
            (timestamp[0] - timestamp[0][0]) / clk_rate,
            None,
            burst["value"].shape,
        )

    @staticmethod
    def _process_node_data(
        node: str, data: t.List[t.Dict[str, t.Any]], clk_rate: float
    ) -> t.List[t.Union[t.Dict[str, t.Any], DAQResult]]:
        """Process the data of a node.

        Only subscribed sample nodes are processed. Other nodes (module native nodes)
        are returned in the original format.

        Args:
            node: Name of the node of the burst.
            data: raw data for the node.
            clk_rate: Clock rate [Hz] for converting the timestamps. Only
                applies if the raw flag is reset.

        Returns:
                Processed and formatted node data.
        """
        if isinstance(data[0], dict):
            return [DAQModule._process_burst(node, burst, clk_rate) for burst in data]
        return data

    def finish(self) -> None:
        """Stop the module.

        .. versionadded:: 0.5.0
        """
        self._raw_module.finish()

    def finished(self) -> bool:
        """Check if the acquisition has finished.

        Returns:
            Flag if the acquisition has finished.

        .. versionadded:: 0.5.0
        """
        return self._raw_module.finished()

    def trigger(self) -> None:
        """Execute a manual trigger.

        .. versionadded:: 0.5.0
        """
        self._raw_module.trigger()

    def read(self, *, raw: bool = False, clk_rate: float = 60e6) -> NodeDict:
        """Read the acquired data from the module.

        The data is split into bursts.

        Args:
            raw: Flag if the acquired data from the subscribed device
                device nodes should be converted into the DAQResult format
                (raw = False) or not. (default = False)
            clk_rate: Clock rate [Hz] for converting the timestamps. Only
                applies if the raw flag is reset.

        Returns:
            Result of the burst grouped by the signals.
        """
        raw_result = self._raw_module.read(flat=True)
        if raw:
            return NodeDict(raw_result)
        return NodeDict(
            {
                node: self._process_node_data(node, data, clk_rate)
                for node, data in raw_result.items()
            }
        )



"""Device Settings Module."""

import logging
import typing as t
from pathlib import Path

from zhinst.core import DeviceSettingsModule as ZIDeviceSettingsModule

from zhinst.toolkit.driver.modules.base_module import BaseModule

from zhinst.toolkit.nodetree.helper import NodeDict

if t.TYPE_CHECKING:  # pragma: no cover
    from zhinst.toolkit.driver.devices import DeviceType
    from zhinst.toolkit.session import Session

logger = logging.getLogger(__name__)


class DeviceSettingsModule(BaseModule):
    """Implements the device settings module for storing and loading settings.

    The Device Settings Module provides functionality for saving and loading
    device settings to and from file. The file is saved in XML format.

    For simple save and load two helper functions exist `save_to_file` and
    `load_from_file`.

    Note: It is not recommend to use this function to read the
        device settings. Instead one can use the zhinst-toolkit functionality
        to read all settings from a device/subtree from the device directly by
        calling it.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/device_settings_module.html

    Args:
        device_settings_module: Instance of the core Impedance Module.
        session: Session to the Data Server.
    """

    def __init__(
        self, device_settings_module: ZIDeviceSettingsModule, session: "Session"
    ):
        super().__init__(device_settings_module, session)
        self.root.update_nodes(
            {
                "/path": {
                    "SetParser": self._set_path,
                }
            },
            raise_for_invalid_node=False,
        )

    def _simple_execution(
        self,
        command: str,
        filename: t.Union[str, Path],
        device: t.Union[str, "DeviceType"],
        timeout: float = 30,
    ) -> None:
        """Execute a command on a clean module.

        This function creates an new module instance to avoid misconfiguration.
        It is also synchronous, meaning it will block until command has
        finished.

        Args:
            command:  The command to execute. (`save`, `load`, `read`)
            filename: The path to the settings file.
            device: The device to load the settings to.
            timeout: Max time to wait for the loading to finish.

        Raises:
            TimeoutError: If the loading of the settings timed out.
        """
        filename = Path(filename)
        temp_module = self._session.modules.create_device_settings_module()
        temp_module.device(device)
        temp_module.filename(filename.stem)
        temp_module.path(filename.parent)
        temp_module.command(command)
        temp_module.execute()
        try:
            # Use finish node instead of function to make advantage of the
            # wait for state change functionality.
            temp_module.finished.wait_for_state_change(1, timeout=timeout)
        except TimeoutError as e:
            raise TimeoutError(
                f"Unable to load device settings after {timeout} seconds."
            ) from e

    def load_from_file(
        self,
        filename: t.Union[str, Path],
        device: t.Union["DeviceType", str],
        timeout: float = 30,
    ) -> None:
        """Load a LabOne settings file to a device.

        This function creates an new module instance to avoid misconfiguration.
        It is also synchronous, meaning it will block until loading the
        settings has finished.

        Args:
            filename: The path to the settings file.
            device: The device to load the settings to.
            timeout: Max time to wait for the loading to finish.

        Raises:
            TimeoutError: If the loading of the settings timed out.
        """
        self._simple_execution("load", filename, device, timeout)

    def save_to_file(
        self,
        filename: t.Union[str, Path],
        device: t.Union["DeviceType", str],
        timeout: int = 30,
    ) -> None:
        """Save the device settings to a LabOne settings file.

        This function creates an new module instance to avoid misconfiguration.
        It is also synchronous, meaning it will block until save operation has
        finished.

        Args:
            filename: The path to the settings file.
            device: The device which settings should be saved.
            timeout: Max time to wait for the loading to finish.

        Raises:
            TimeoutError: If the loading of the settings timed out.
        """
        self._simple_execution("save", filename, device, timeout)

    def read(self) -> NodeDict:
        """Read device settings.

        Note: It is not recommend to use this function to read the
        device settings. Instead one can use the zhinst-toolkit functionality
        to read all settings from a device/subtree from the device directly by
        calling it.

        >>> device = session.connect_device()
        >>> ...
        >>> device()
        <all device settings>
        >>> device.demods()
        <all demodulator settings>

        Returns:
            Device settings.
        """
        return super().read()



"""Base Module Driver.

Natively works with all module types and provides the basic functionality like
the module specific nodetree.
"""
import logging
import time
import typing as t
from os import PathLike, fspath
from functools import partial

from zhinst.core import ModuleBase
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.helper import NodeDict

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:  # pragma: no cover
    from zhinst.toolkit.driver.devices import DeviceType
    from zhinst.toolkit.session import Session

ZIModule = t.TypeVar("ZIModule", bound=ModuleBase)


class BaseModule(Node):
    """Generic toolkit driver for a LabOne Modules.

    All module specific class are derived from this class.
    It exposes the nodetree and also implements common functions valid for all
    modules.
    It also can be used directly, e.g. for modules that have no special class
    in toolkit.

    Args:
        raw_module: zhinst.core module.
        session: Session to the Data Server.
    """

    def __init__(self, raw_module: ZIModule, session: "Session"):
        self._raw_module = raw_module
        self._session = session
        super().__init__(NodeTree(raw_module), tuple())
        self.root.update_nodes(
            {
                "/device": {
                    "GetParser": partial(self._get_device, self._session),
                    "SetParser": self._set_device,
                },
                "/directory": {
                    "SetParser": self._set_path,
                },
                "/save/directory": {
                    "SetParser": self._set_path,
                },
            },
            raise_for_invalid_node=False,
        )

    def __repr__(self):
        return str(f"{self._raw_module.__class__.__name__}({repr(self._session)})")

    @staticmethod
    def _get_device(session, serial: str) -> t.Union["DeviceType", str]:
        """Convert a device serial into a toolkit device object.

        Args:
            serial: Serial of the device

        Returns:
            Toolkit device object. If the serial does not
                match to a connected device the serial is returned instead.
        """
        try:
            return session.devices[serial]
        except (RuntimeError, KeyError):
            return serial

    @staticmethod
    def _set_device(value: t.Union["DeviceType", str]) -> str:
        """Convert a toolkit device object into a serial string.

        Args:
            serial: A toolkit device object
                (can also be a serial string directly)

        Returns:
            str: device serial
        """
        try:
            return value.serial  # type: ignore
        except AttributeError:
            return value

    def _get_node(self, node: str) -> t.Union[Node, str]:
        """Convert a raw node string into a toolkit node.

        Args:
            node (str): raw node string

        Returns:
            Toolkit node. (if the node can not be converted the raw node
                string is returned)
        """
        try:
            return self._session.raw_path_to_node(node.replace(".", "/"), module=self)
        except (KeyError, RuntimeError):
            logger.error(
                f"Could not resolve {node} into a node of the "
                f"{self._raw_module.__class__.__name__} or "
                "a connected device."
            )
            return node

    @staticmethod
    def _set_node(signal: t.Union[Node, str]) -> str:
        """Convert a toolkit node into a raw node string.

        Args:
            signal: Toolkit node

        Returns:
            str: raw string node
        """
        try:
            node = signal.node_info.path  # type: ignore
        except AttributeError:
            node = signal
        return node

    @staticmethod
    def _set_path(path: t.Union[str, PathLike]) -> str:
        """Convert a Pathlike object into a string for LabOne.

        LabOne only accepts strings for the path node. This function converts
        any Pathlike object into a string so that is processed by LabOne
        correctly.

        If the object implements the `absolute` method (e.g. pathlib.Path)
        it is used to use the absolute path. This is important since LabOne
        operates in a different directory.

        Args:
            path (str): Path to send to the device.

        Returns:
            String representation of the path like object.
        """
        if hasattr(path, "absolute"):
            path = path.absolute()  # type: ignore[union-attr]
        return fspath(path)

    def wait_done(self, *, timeout: float = 20.0, sleep_time: float = 0.5) -> None:
        """Waits until the module is finished.

        Warning: Only usable for modules that make use of the `/finished` node.

        Args:
            timeout (float): The maximum waiting time in seconds for the
                measurement (default: 20).
            sleep_time (int): Time in seconds to wait between
                requesting sweeper state. (default: 0.5)

        Raises:
            TimeoutError: The measurement is not completed before
                timeout.
        """
        start_time = time.time()
        while (
            start_time + timeout >= time.time()
            and not self._raw_module.finished()
            and self.progress() != 1
        ):
            logger.info(f"Progress: {(self.progress() * 100):.1f}%")
            time.sleep(sleep_time)
        if not self._raw_module.finished() and self.progress() != 1:
            raise TimeoutError(f"{self._raw_module.__class__.__name__} timed out.")
        logger.info(f"Progress: {(self.progress() * 100):.1f}%")

    def progress(self) -> float:
        """Progress of the execution.

        Returns:
            Progress of the execution with a number between 0 and 1
        """
        return self._raw_module.progress()[0]

    def subscribe(self, signal: t.Union[Node, str]) -> None:
        """Subscribe to a node.

        The node can either be a node of this module or of a connected device.

        Args:
            signal: Node that should be subscribed to.

        .. versionchanged 0.5.0 Add support for raw string signals
        """
        try:
            self._raw_module.subscribe(signal.node_info.path)  # type: ignore
        except AttributeError:
            self._raw_module.subscribe(signal)

    def unsubscribe(self, signal: t.Union[Node, str]) -> None:
        """Unsubscribe from a node.

        The node can either be a node of this module or of a connected device.

        Args:
            signal: Node that should be unsubscribed from.

        .. versionchanged 0.5.0 Add support for raw string signals
        """
        try:
            self._raw_module.unsubscribe(signal.node_info.path)  # type: ignore
        except AttributeError:
            self._raw_module.unsubscribe(signal)

    def execute(self) -> None:
        """Start the module execution.

        Subscription or unsubscription is not possible until the execution is
        finished.

        .. versionadded:: 0.4.1
        """
        self._raw_module.execute()

    def read(self) -> NodeDict:
        """Read scope data.

        If the recording is still ongoing only a subset of data is returned.

        Returns:
            Scope data.
        """
        return NodeDict(self._raw_module.read(flat=True))

    @property
    def raw_module(self) -> ZIModule:  # type: ignore [type-var]
        """Underlying core module."""
        return self._raw_module



import pytest

from zhinst.toolkit.exceptions import ToolkitError


def test_base_exception():
    assert issubclass(ToolkitError, RuntimeError)


def test_base_exception_raises():
    with pytest.raises(RuntimeError):
        raise ToolkitError("message")



from zhinst.toolkit import Session
from zhinst.toolkit.driver.nodes.awg import AWG
import enum
from typing import Optional
from dataclasses import dataclass
import numpy as np


class LogType(enum.Enum):
    Error = enum.auto()
    Trigger = enum.auto()
    ZSync_Feedback = enum.auto()
    Internal_Feedback = enum.auto()
    ZSync_AUX = enum.auto()
    DIO = enum.auto()


class TriggerSource(enum.Flag):
    DigTrigger1 = enum.auto()
    DigTrigger2 = enum.auto()
    ZSyncTrigger = enum.auto()


@dataclass
class LogEntry:
    time_clk: int
    time_us: float
    log_type: LogType
    trigger_source: Optional[TriggerSource] = None
    raw: Optional[int] = None
    processed1: Optional[int] = None
    processed2: Optional[int] = None
    addr: Optional[int] = None
    data: Optional[int] = None

    def __str__(self):
        entry_str = f"{self.time_clk:<7d}\t{self.time_us:.3f}\t{self.log_type.name:s} "

        if self.log_type == LogType.Error:
            entry_str += "Collision error"
        if self.log_type == LogType.Trigger:
            entry_str += f"source({self.trigger_source.name:s})"
        elif (
            self.log_type == LogType.ZSync_Feedback
            and self.processed1 is not None
            and self.processed2 is not None
        ):
            entry_str += f"raw(0x{self.raw:04X}) register(0x{self.processed1:04X}) decoder(0x{self.processed2:04X})"
        elif self.log_type == LogType.ZSync_Feedback:
            entry_str += f"raw(0x{self.raw:04X})"
        elif self.log_type == LogType.Internal_Feedback:
            entry_str += f"raw(0x{self.raw:04X}) processed(0x{self.processed1:04X})"
        elif self.log_type == LogType.ZSync_AUX:
            entry_str += f"addr(0x{self.addr:04X}) data(0x{self.data:08X})"
        elif self.log_type == LogType.DIO:
            entry_str += f"raw(0x{self.raw:08X})"
        else:
            raise RuntimeError("Unknown log type!")

        return entry_str


def reset_and_enable_rtlogger(
    awg: AWG, input: str = "zsync", start_timestamp: Optional[int] = None
) -> None:
    """Reset and start a given RT Logger.

    Args:
        awg: AWG node of the RTLogger
        input: The source of data that it should log. Either "dio" or "zsync". (default: "zsync")
        start_timestamp: optional start timestamp, if provided, timestamp mode is used.

    """
    awg.rtlogger.clear(True)  # Clear the logs
    if start_timestamp is None:
        # Starts with the AWG and overwrites old values as soon as the memory limit
        # is reached.
        awg.rtlogger.mode("normal")
    else:
        # Starts with the AWG, waits for the first valid trigger, and only starts
        # recording data after the time specified by the start_timestamp.
        # Recording stops as soon as the memory limit is reached.
        awg.rtlogger.mode("timestamp")
        awg.rtlogger.starttimestamp(start_timestamp)

    # Set the input of the rtlogger
    # This is necessary only on the SHF family,
    # on the HDAWG such node is absent, since the input
    # is selected by the node dios/0/mode
    if awg.rtlogger.input:
        awg.rtlogger.input(input)

    # Start the rtlogger
    awg.rtlogger.enable(True, deep=True)


def _get_trigdelay(session: Session, awg: AWG) -> int:
    """Get the ZSync trigger delay.

    Note: this function makes use of a raw node; raw nodes are usually meant for
    internal purposes only, they are not documented and their existence is not
    guaranteed in future releases. This function is intended for illustrative
    purposes only and raw nodes should not be used by users.

    Args:
        session: Toolkit session to a data server.
        awg: AWG node.

    Returns:
        The delay of the ZSync trigger, in clock cycles.

    """
    awg_split = str(awg).split("/")  # Split into list
    awg_split.insert(2, "raw")
    awg_split.append("zsync")
    awg_split.append("trigdelay")
    trigdelay_node = "/".join(awg_split)  # Join again into node path

    trig_delay = session.daq_server.getInt(trigdelay_node)

    # The trigger delay on the SG and QA channels is in 500 MHz unit
    if "sgchannels" in str(awg) or "qachannels" in str(awg):
        trig_delay = int(np.ceil(trig_delay / 2))

    return trig_delay


def print_rtlogger_data(
    session: Session,
    awg: AWG,
    compensate_start_trigger: bool = True,
    max_lines: int = None,
    silent: bool = False,
) -> Optional[list[LogEntry]]:
    """Print the data collected by the RT Logger.

    Args:
        session: Toolkit session to a data server.
        awg: AWG node
        compensate_start_trigger: True if the start trigger delay should
            be compensated for.
        max_lines: Maximum number of lines to be printed.
        silent: if True - don't print anything, return the
            list of events. (Default: False)

    Returns:
        list[LogEntry]: list of logged events, if silent is True, otherwise None

    """
    rtlogger = awg.rtlogger

    # Fetch the output of the rtlogger and decode
    rtdata = rtlogger.data().reshape((-1, 4))
    rtdata.dtype = np.dtype(
        [
            ("timestamp", np.int64),
            ("value", np.int64),
            ("source", np.int64),
            ("error", np.int64),
        ]
    )

    # Get various parameter
    timebase = rtlogger.timebase()

    reg_node = awg.zsync.register
    if reg_node:
        reg_shift, reg_mask, reg_offset = (
            reg_node.shift(),
            reg_node.mask(),
            reg_node.offset(),
        )
    else:
        reg_shift, reg_mask, reg_offset = (0, 0, 0)

    dec_node = awg.zsync.decoder
    if dec_node:
        dec_shift, dec_mask, dec_offset = (
            dec_node.shift(),
            dec_node.mask(),
            dec_node.offset(),
        )
    else:
        dec_shift, dec_mask, dec_offset = (0, 0, 0)

    intfeedback_node = awg.intfeedback.direct
    if intfeedback_node:
        int_shift, int_mask, int_offset = (
            intfeedback_node.shift(),
            intfeedback_node.mask(),
            intfeedback_node.offset(),
        )
    else:
        int_shift, int_mask, int_offset = (0, 0, 0)

    if compensate_start_trigger:
        trig_delay = _get_trigdelay(session, awg)
    else:
        trig_delay = 0
    base_ts = 0

    # Process the raw data
    max_lines = max_lines or len(rtdata)
    entries = []
    for i in range(max_lines):
        line = rtdata[i]
        raw_value = int(line["value"])

        # the +2 is due to the difference between the
        # rtlogger and the sequencer behavior
        ts = int(line["timestamp"]) - base_ts + 2
        ts_s = ts * timebase * 1e6

        # Check collision error
        if line["error"]:
            entry = LogEntry(ts, ts_s, LogType.Error)
            entries.append(entry)
            continue

        # - Trigger processing
        trigger_source = TriggerSource(0)
        if line["source"] == 2:
            # Dig trigger 1
            if raw_value & 0x40000000:
                trigger_source |= TriggerSource.DigTrigger1
            # Dig trigger 2
            if raw_value & 0x80000000:
                trigger_source |= TriggerSource.DigTrigger2
            # ZSync trigger
            if (raw_value & 0xC0000000) == 0 and (raw_value & 0xFF) == 0x08:
                trigger_source = TriggerSource.ZSyncTrigger

        if bool(trigger_source):
            # Reset the time counter if requested
            if compensate_start_trigger:
                base_ts = int(line["timestamp"]) + trig_delay
                ts = 0
                ts_s = 0.0
            else:
                ts = int(line["timestamp"])
                ts_s = ts * timebase * 1e6

            # We got a trigger, save it
            entry = LogEntry(ts, ts_s, LogType.Trigger, trigger_source=trigger_source)
            entries.append(entry)
            continue

        # - ZSync feedback processing
        if line["source"] == 1:
            if reg_node and dec_node:
                register_data = ((raw_value >> reg_shift) & reg_mask) + reg_offset
                decoder_data = ((raw_value >> dec_shift) & dec_mask) + dec_offset
                entry = LogEntry(
                    ts,
                    ts_s,
                    LogType.ZSync_Feedback,
                    raw=raw_value,
                    processed1=register_data,
                    processed2=decoder_data,
                )
            else:
                entry = LogEntry(ts, ts_s, LogType.ZSync_Feedback, raw=raw_value)
            entries.append(entry)
            continue

        # - Internal feedback processing
        if line["source"] == 3:
            processed_data = ((raw_value >> int_shift) & int_mask) + int_offset
            entry = LogEntry(
                ts,
                ts_s,
                LogType.Internal_Feedback,
                raw=raw_value,
                processed1=processed_data,
            )
            entries.append(entry)
            continue

        # - DIO processing
        if line["source"] == 0:
            entry = LogEntry(ts, ts_s, LogType.DIO, raw=raw_value)
            entries.append(entry)
            continue

        # - ZSync AUX processing
        if (
            line["source"] == 2
            and (raw_value & 0xC0000000) == 0
            and (raw_value & 0xFF) == 0x01
        ):
            addr = (raw_value >> 8) & 0xFFFF
            data = (raw_value >> 16) & 0x3FFF
            entry = LogEntry(ts, ts_s, LogType.ZSync_AUX, addr=addr, data=data)
            entries.append(entry)
            continue

    if silent:
        return entries
    else:
        # Print the RTLogger logs
        print("t[clk]\tt[us]\tData")
        for entry in entries:
            print(entry)
        return None  # Avoid dump to console in interactive session



"""Helper functions used in toolkit."""
import typing as t
from contextlib import contextmanager
from collections.abc import Mapping
import re
from _thread import RLock  # type: ignore
from functools import wraps

# TypedDict is available in the typing module since 3.8
# Ift we only support 3.8 we should switch to t.TypedDict
from typing_extensions import TypedDict

if t.TYPE_CHECKING:  # pragma: no cover
    from zhinst.toolkit.nodetree.node import Node

T = t.TypeVar("T")

_NodeInfo = TypedDict(
    "_NodeInfo",
    {
        "Node": str,
        "Description": str,
        "Properties": str,
        "Type": str,
        "Unit": str,
        "Options": t.Dict[str, str],
    },
)
NodeDoc = t.Dict[str, _NodeInfo]

_NOT_FOUND = object()


# Exact implementation of functools.cached_property from Python 3.8
# This is needed for Python 3.7 compatibility
# It should be removed once we drop support for Python 3.7
class lazy_property:
    """Copied functools.cached_property from Python 3.8.

    Decorator that converts a method with a single self argument into a
    property cached on the instance.
    """

    def __init__(self, func):
        self.func = func
        self.attrname = None
        self.__doc__ = func.__doc__
        self.lock = RLock()

    def __set_name__(self, owner, name):
        if self.attrname is None:
            self.attrname = name
        elif name != self.attrname:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                f"({self.attrname!r} and {name!r})."
            )

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError(
                "Cannot use cached_property instance without "
                "calling __set_name__ on it."
            )
        try:
            cache = instance.__dict__
        except AttributeError:  # not all objects have __dict__
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {self.attrname!r} property."
            )
            raise TypeError(msg) from None
        val = cache.get(self.attrname, _NOT_FOUND)
        if val is _NOT_FOUND:
            with self.lock:
                # check if another thread filled cache while we awaited lock
                val = cache.get(self.attrname, _NOT_FOUND)
                if val is _NOT_FOUND:
                    val = self.func(instance)
                    try:
                        cache[self.attrname] = val
                    except TypeError:
                        msg = (
                            f"The '__dict__' attribute on {type(instance).__name__!r} "
                            "instance does not support item assignment for caching "
                            f"{self.attrname!r} property."
                        )
                        raise TypeError(msg) from None
        return val


@contextmanager
def create_or_append_set_transaction(nodetree) -> t.Generator[None, None, None]:
    """Context manager for a transactional set.

    In contrast to the set_transaction from the nodetree this function only
    creates a new transaction if no other is in progress.
    Should only be called withing the toolkit code.

    Warning:
        This function will silently fail if the existing transaction is exited
        before this function finishes.

    Warning:
        The set is always performed as deep set if called on device nodes.

    Examples:
        >>> with nodetree.set_transaction():
                nodetree.test[0].a(1)
                with create_or_append_set_transaction(nodetree):
                    nodetree.test[1].a(2)
                nodetree.test[2].a(2)
    """
    if not nodetree.transaction.in_progress():
        with nodetree.set_transaction():
            yield
    else:
        yield


def resolve_wildcards_labone(path: str, nodes: t.List[str]) -> t.List[str]:
    """Resolves potential wildcards.

    Also will resolve partial nodes to its leaf nodes.

    Returns:
        List of matched nodes in the raw path format
    """
    node_raw = re.escape(path)
    node_raw = node_raw.replace("/\\*/", "/[^/]*/").replace("/\\*", "/*") + "(/.*)?$"
    node_raw_regex = re.compile(node_raw)
    return list(filter(node_raw_regex.match, nodes))


class NodeDict(Mapping):
    """Mapping of dictionary structure results.

    The mapping allows to access data with both the string and the toolkit
    node objects.

    Args:
        result: A dictionary of node/value pairs.

    Example:
        >>> result = device.demods["*"].enable()
        >>> print(result)
        {
            '/dev1234/demods/0/enable': 0,
            '/dev1234/demods/1/enable': 1,
        }
        >>> result[device.demods[0].enable]
        0
        >>> result["/dev1234/demods/0/enable"]
        0

    .. versionadded:: 0.3.5 Renamed from WildcardResult
    """

    def __init__(self, result: t.Dict[str, t.Any]):
        self._result = result

    def __repr__(self):
        return repr(self._result)

    def __getitem__(self, key: t.Union[str, "Node"]):
        return self._result[str(key)]

    def __iter__(self):
        return iter(self._result)

    def __len__(self):
        return len(self._result)

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Convert the WildcardResult to a dictionary.

        After conversion, :class:`Node` objects cannot be used to get items.
        """
        return self._result


def not_callable_in_transactions(
    func: t.Callable[["Node", t.Any], t.Any]
) -> t.Callable[["Node", t.Any], t.Any]:
    """Wrapper to prevent certain functions from being used within a transaction.

    Certain utils functions which that both get and set values would not work like
    expected in a transaction. This wrapper prevents misuse by throwing an error
    in such cases.

    Args:
        func: function to wrap

    Returns:
        Similar function, but not callable from transactions
    """

    @wraps(func)
    def wrapper(node: "Node", *args, **kwargs):
        if node.root.transaction.in_progress():
            raise RuntimeError(
                f"'{func.__name__}' cannot be called inside a transaction"
            )
        return func(node, *args, **kwargs)

    return wrapper



"""PID Advisor Module."""
import logging
import typing as t
import time
from enum import IntFlag

from zhinst.core import PidAdvisorModule as ZIPidAdvisorModule

from zhinst.toolkit.driver.modules.base_module import BaseModule

if t.TYPE_CHECKING:  # pragma: no cover
    from zhinst.toolkit.session import Session

logger = logging.getLogger(__name__)


class PIDMode(IntFlag):
    """PID Advisor mode.

    P_Gain:         Optimize/Tune P gain.
    I_Gain:         Optimize/Tune I gain.
    D_Gain:         Optimize/Tune D gain.
    D_Filter_Limit: Optimize/Tune D filter limit.

    .. versionadded:: 0.5.1
    """

    NONE = 0
    P_Gain = 1
    I_Gain = 2
    D_Gain = 4
    D_Filter_Limit = 8


class PIDAdvisorModule(BaseModule):
    """PID Advisor Module.

    The PID Advisor Module provides the functionality available in the Advisor,
    Tuner and Display sub-tabs of the LabOne User Interface’s PID / PLL tab.
    The PID Advisor is a mathematical model of the instrument’s PID and can be
    used to calculate PID controller parameters for optimal feedback loop
    performance. The controller gains calculated by the module can be easily
    transferred to the device via the API and the results of the Advisor’s
    modeling are available as Bode and step response plot data.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/pid_advisor_module.html

    Args:
        pid_advisor_module: Instance of the core PID advisor module.
        session: Session to the Data Server.

    .. versionadded:: 0.5.1
    """

    def __init__(self, pid_advisor_module: ZIPidAdvisorModule, session: "Session"):
        super().__init__(pid_advisor_module, session)
        self.root.update_nodes(
            {
                "/pid/mode": {
                    "GetParser": PIDMode,
                },
                "/tuner/mode": {
                    "GetParser": PIDMode,
                },
            },
            raise_for_invalid_node=False,
        )

    def wait_done(self, *, timeout: float = 20.0, sleep_time: float = 2) -> None:
        """Waits until the pid advisor is finished.

        Args:
            timeout (float): The maximum waiting time in seconds for the
                measurement (default: 20).
            sleep_time (int): Time in seconds to wait between
                requesting sweeper state. (default: 0.5)

        Raises:
            TimeoutError: The measurement is not completed before
                timeout.
        """
        start_time = time.time()
        while start_time + timeout >= time.time() and (
            self.calculate()
            or (not self._raw_module.finished() and self.progress() != 1)
        ):
            # When the advisor is started it takes some time to reset the
            # progress (and finished) node. This causes weird behavior when
            # restarted. This if statement just ignores the progress until it
            # is smaller than 100%
            if self.progress() < 1:
                logger.info(f"Progress: {(self.progress() * 100):.1f}%")
            else:
                logger.info("Progress: 0.0%")
            time.sleep(sleep_time)
        if self.calculate() or (
            not self._raw_module.finished() and self.progress() != 1
        ):
            raise TimeoutError(f"{self._raw_module.__class__.__name__} timed out.")
        logger.info(f"Progress: {(self.progress() * 100):.1f}%")



"""Global zhinst-toolkit exceptions."""


class ToolkitError(RuntimeError):
    """Base class for `zhinst.toolkit` errors.

    .. versionadded:: 0.5.1
    """


class ValidationError(ToolkitError):
    """Data validation failed.

    .. versionchanged:: 0.5.1

         Changed base class from `Exception` to `ToolkitError`.
    """



from itertools import cycle
import pytest

from zhinst.toolkit.driver.devices.pqsc import PQSC
from zhinst.toolkit.exceptions import ToolkitError


@pytest.fixture()
def pqsc(data_dir, mock_connection, session):

    json_path = data_dir / "nodedoc_dev1234_pqsc.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json
    mock_connection.return_value.getString.return_value = ""

    yield PQSC("DEV1234", "PQSC", session)


def test_repr(pqsc):
    assert repr(pqsc) == "PQSC(PQSC,DEV1234)"


def test_arm(mock_connection, pqsc):
    pqsc.arm()
    mock_connection.return_value.syncSetInt.assert_any_call(
        "/dev1234/execution/enable", False
    )
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/feedback/registerbank/reset", 1
    )

    pqsc.arm(repetitions=3, holdoff=100.6)
    mock_connection.return_value.syncSetInt.assert_any_call(
        "/dev1234/execution/enable", False
    )
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/repetitions", 3
    )
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/holdoff", 100.6
    )
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/feedback/registerbank/reset", 1
    )

    pqsc.arm(deep=False, repetitions=5, holdoff=67.4)
    mock_connection.return_value.set.assert_any_call("/dev1234/execution/enable", False)
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/repetitions", 5
    )
    mock_connection.return_value.set.assert_any_call("/dev1234/execution/holdoff", 67.4)
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/feedback/registerbank/reset", 1
    )


def test_run_stop(mock_connection, pqsc):
    pqsc.run(deep=False)
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/execution/enable", True
    )
    pqsc.run()
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/execution/enable", True
    )

    pqsc.stop(deep=False)
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/execution/enable", False
    )
    pqsc.stop()
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/execution/enable", False
    )
    mock_connection.reset_mock()

    pqsc.arm_and_run(repetitions=5, holdoff=100.8)
    mock_connection.return_value.syncSetInt.assert_any_call(
        "/dev1234/execution/enable", False
    )
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/repetitions", 5
    )
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/holdoff", 100.8
    )
    mock_connection.return_value.syncSetInt.assert_any_call(
        "/dev1234/feedback/registerbank/reset", 1
    )
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/execution/enable", True
    )


def test_wait_done(mock_connection, pqsc):
    sequence = iter([1] * 3 + [0] * 5)
    mock_connection.return_value.getInt.side_effect = lambda node: next(sequence)
    mock_connection.return_value.get.side_effect = lambda node, **kwargs: {
        node: {"timestamp": [0], "value": [next(sequence)]}
    }
    pqsc.wait_done(sleep_time=0.001)
    mock_connection.return_value.getInt.assert_called_with("/dev1234/execution/enable")

    mock_connection.return_value.getInt.side_effect = lambda node: 1
    mock_connection.return_value.get.side_effect = lambda node, **kwargs: {
        node: {"timestamp": [0], "value": [1]}
    }
    with pytest.raises(TimeoutError) as e_info:
        pqsc.wait_done(timeout=0.01)
    mock_connection.return_value.getInt.assert_called_with("/dev1234/execution/enable")


def test_ref_clock(mock_connection, pqsc):

    status = cycle([0])
    source = 0
    source_actual = 0

    def getInt_side_effect(path):
        if path == "/dev1234/system/clocks/referenceclock/in/status":
            return next(status)
        if path == "/dev1234/system/clocks/referenceclock/in/source":
            return source
        if path == "/dev1234/system/clocks/referenceclock/in/sourceactual":
            return source_actual
        raise RuntimeError("Invalid Node")

    def get_side_effect(path, **kwargs):
        value = getInt_side_effect(path)
        return {path: {"timestamp": [0], "value": [value]}}

    mock_connection.return_value.getInt.side_effect = getInt_side_effect
    mock_connection.return_value.get.side_effect = get_side_effect

    assert pqsc.check_ref_clock(sleep_time=0.001)
    # Locked within time
    status = iter([2] * 2 + [0] * 10)
    assert pqsc.check_ref_clock(sleep_time=0.001)
    # Locking error but actual_clock == clock
    status = cycle([1])
    assert not pqsc.check_ref_clock(sleep_time=0.001)
    # Locking error and actual_clock != clock => reset clock to internal
    source = 1
    mock_connection.return_value.syncSetString.assert_not_called()
    assert not pqsc.check_ref_clock(sleep_time=0.001)
    mock_connection.return_value.syncSetString.assert_called_with(
        "/dev1234/system/clocks/referenceclock/in/source", "internal"
    )

    # timeout
    status = cycle([2])
    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_ref_clock(timeout=0.01, sleep_time=0.001)


def test_check_zsync_connection(mock_connection, pqsc, shfqa, shfsg, shfqc):

    # hijack serials of instruments
    shfsg._serial = "dev1111"  # port 0
    shfqa._serial = "dev2222"  # port 2
    shfqc._serial = "dev3333"  # port 10

    status0 = 0
    status2 = 0
    status10 = 0

    def getInt_side_effect(path):
        if path == "/dev1234/zsyncs/0/connection/status":
            return status0
        if path == "/dev1234/zsyncs/2/connection/status":
            return status2
        if path == "/dev1234/zsyncs/10/connection/status":
            return status10
        raise RuntimeError("Invalid Node")

    def get_serials(path):
        # assume it ask for path "/dev1234/zsyncs/*/connection/serial"
        return_dict = {}
        for port in range(18):
            key = f"/dev1234/zsyncs/{port:d}/connection/serial"
            if port == 0 and status0 != 0:
                serial = "1111"
            elif port == 2 and status2 != 0:
                serial = "2222"
            elif port == 10 and status10 != 0:
                serial = "3333"
            else:
                serial = ""

            return_dict[key] = [{"timestamp": 0, "flags": 0, "vector": serial}]

        return return_dict

    def get_side_effect(path, **kwargs):
        if path.endswith("status"):
            value = getInt_side_effect(path)
            return {path: {"timestamp": [0], "value": [value]}}
        elif path.endswith("serial"):
            return get_serials(path)
        else:
            return None

    mock_connection.return_value.getInt.side_effect = getInt_side_effect
    mock_connection.return_value.get.side_effect = get_side_effect

    # not connected
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection(0, timeout=0, sleep_time=0.001)
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection(2, timeout=0, sleep_time=0.001)
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection([0], timeout=0, sleep_time=0.001)
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection([0, 2], timeout=0, sleep_time=0.001)

    # one connected
    status0 = 2
    assert True == pqsc.check_zsync_connection(0)
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection(2, timeout=0, sleep_time=0.001)
    assert [True] == pqsc.check_zsync_connection([0])
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection([0, 2], timeout=0, sleep_time=0.001)

    # one connected the other one in error state
    status0 = 2
    status2 = 3
    assert True == pqsc.check_zsync_connection(0, sleep_time=0.001)
    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_zsync_connection(2, timeout=0.01)
    assert [True] == pqsc.check_zsync_connection([0], sleep_time=0.001)

    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_zsync_connection([0, 2], timeout=0.01, sleep_time=0.001)
    assert "2" in str(e_info.value)

    # one connected the other one times out
    status0 = 2
    status2 = 1
    assert True == pqsc.check_zsync_connection(0, sleep_time=0.001)
    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_zsync_connection(2, timeout=0.01)
    assert [True] == pqsc.check_zsync_connection([0], sleep_time=0.001)

    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_zsync_connection([0, 2], timeout=0.01, sleep_time=0.001)
    assert "2" in str(e_info.value)

    # test two decimal place value
    status10 = 2
    status0 = 2
    status2 = 3
    assert [True] == pqsc.check_zsync_connection([10], sleep_time=0.001)
    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_zsync_connection([0, 2, 10], timeout=0.01, sleep_time=0.001)
    assert "2" in str(e_info.value)

    # Test with instruments
    status0 = 2
    status2 = 2

    assert True == pqsc.check_zsync_connection(shfsg, timeout=0.05, sleep_time=0.001)
    assert True == pqsc.check_zsync_connection(shfqa, timeout=0.05, sleep_time=0.001)
    assert [True, True] == pqsc.check_zsync_connection(
        [shfsg, shfqa], timeout=0.05, sleep_time=0.001
    )

    status10 = 3
    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_zsync_connection(shfqc, timeout=0.05, sleep_time=0.001)

    status10 = 0
    with pytest.raises(ToolkitError) as e_info:
        pqsc.check_zsync_connection(shfqc, timeout=0.05, sleep_time=0.001)


def test_find_zsync_worker_port(mock_connection, pqsc):
    # device not connected
    ret_val = {"random/path": [{"timestamp": 0, "flags": 0, "vector": ""}]}
    mock_connection.return_value.get.return_value = ret_val
    with pytest.raises(RuntimeError):
        pqsc.find_zsync_worker_port(pqsc)

    # correct use case
    ret_val = {
        "/dev1234/zsyncs/0/connection/serial": [
            {"timestamp": 0, "flags": 0, "vector": "1234"}
        ]
    }
    mock_connection.return_value.get.return_value = ret_val
    assert pqsc.find_zsync_worker_port(pqsc) == 0



from unittest.mock import patch

import pytest

from zhinst.toolkit.driver.devices.shfsg import AWG, Node, SGChannel


def test_repr(shfsg):
    shfsg.sgchannels["*"].configure_channel()
    assert repr(shfsg) == "SHFSG(SHFSG8,DEV1234)"


def test_sgchannels(shfsg, mock_connection):
    assert len(shfsg.sgchannels) == 8
    assert isinstance(shfsg.sgchannels[0], SGChannel)
    # Wildcards nodes will be converted into normal Nodes
    assert not isinstance(shfsg.sgchannels["*"], SGChannel)
    assert isinstance(shfsg.sgchannels["*"], Node)


def test_sg_awg(shfsg):
    assert isinstance(shfsg.sgchannels[0].awg, AWG)
    assert shfsg.sgchannels[0].awg.root == shfsg.root
    assert shfsg.sgchannels[0].awg.raw_tree == shfsg.raw_tree + (
        "sgchannels",
        "0",
        "awg",
    )


def test_sg_awg_compiler(mock_connection, shfsg):
    mock_connection.return_value.getString.return_value = "AWG,FOOBAR"
    with patch(
        "zhinst.toolkit.driver.nodes.awg.compile_seqc", autospec=True
    ) as compile_seqc:
        shfsg.sgchannels[0].awg.compile_sequencer_program("test")
        compile_seqc.assert_called_once_with("test", "SHFSG8", "AWG,FOOBAR", 0)


def test_sg_awg_modulation_freq(mock_connection, shfsg, data_dir):
    mock_connection.return_value.getInt.return_value = 0
    shfsg.sgchannels[0].awg_modulation_freq()
    mock_connection.return_value.getInt.assert_called_with(
        "/dev1234/sgchannels/0/sines/0/oscselect"
    )
    mock_connection.return_value.getDouble.assert_called_with(
        "/dev1234/sgchannels/0/oscs/0/freq"
    )
    mock_connection.return_value.getInt.return_value = 1
    shfsg.sgchannels[0].awg_modulation_freq()
    mock_connection.return_value.getInt.assert_called_with(
        "/dev1234/sgchannels/0/sines/0/oscselect"
    )
    mock_connection.return_value.getDouble.assert_called_with(
        "/dev1234/sgchannels/0/oscs/1/freq"
    )


def test_awg_configure_marker_and_trigger(data_dir, mock_connection, shfsg):
    with patch("zhinst.toolkit.driver.devices.shfsg.utils", autospec=True) as utils:
        shfsg.sgchannels[0].awg.configure_marker_and_trigger(
            trigger_in_source="test1",
            trigger_in_slope="test2",
            marker_out_source="test3",
        )
        utils.get_marker_and_trigger_settings.assert_called_once_with(
            "DEV1234",
            0,
            trigger_in_source="test1",
            trigger_in_slope="test2",
            marker_out_source="test3",
        )

    assert shfsg.sgchannels[0].awg.available_trigger_inputs == [
        "trigin0",
        "trigin1",
        "trigin2",
        "trigin3",
        "trigout0",
        "trigout1",
        "trigout2",
        "trigout3",
    ]

    assert shfsg.sgchannels[0].awg.available_trigger_slopes == [
        "level_sensitive",
        "rising_edge",
        "falling_edge",
        "both_edges",
    ]

    assert shfsg.sgchannels[0].awg.available_marker_outputs == [
        "awg_trigger0",
        "awg_trigger1",
        "awg_trigger2",
        "awg_trigger3",
        "output0_marker0",
        "output0_marker1",
        "output1_marker0",
        "output1_marker1",
        "trigin0",
        "trigin1",
        "trigin2",
        "trigin3",
        "trigin4",
        "trigin5",
        "trigin6",
        "trigin7",
        "high",
        "low",
    ]


def test_configure_channel(mock_connection, shfsg):
    with patch("zhinst.toolkit.driver.devices.shfsg.utils", autospec=True) as utils:
        shfsg.sgchannels[0].configure_channel(
            enable=True, output_range=111, center_frequency=20.4, rf_path=False
        )
        utils.configure_channel.assert_called_once_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            enable=1,
            output_range=111,
            center_frequency=20.4,
            rflf_path=0,
        )


def test_configure_pulse_modulation(mock_connection, shfsg):
    with patch("zhinst.toolkit.driver.devices.shfsg.utils", autospec=True) as utils:
        shfsg.sgchannels[0].configure_pulse_modulation(
            enable=True,
            osc_index=5,
            osc_frequency=22.0,
            phase=856.6789,
            global_amp=7.0,
            gains=(3.0, -1.0, 5.0, 1.0),
            sine_generator_index=8,
        )
        utils.get_pulse_modulation_settings.assert_called_once_with(
            "DEV1234",
            0,
            enable=1,
            osc_index=5,
            osc_frequency=22.0,
            phase=856.6789,
            global_amp=7.0,
            gains=(3.0, -1.0, 5.0, 1.0),
            sine_generator_index=8,
        )


def test_configure_sine_generation(mock_connection, shfsg):
    with patch("zhinst.toolkit.driver.devices.shfsg.utils", autospec=True) as utils:
        shfsg.sgchannels[0].configure_sine_generation(
            enable=True,
            osc_index=5,
            osc_frequency=22.0,
            phase=856.6789,
            gains=(3.0, -1.0, 5.0, 1.0),
            sine_generator_index=8,
        )
        utils.get_sine_generation_settings.assert_called_once_with(
            "DEV1234",
            0,
            enable=1,
            osc_index=5,
            osc_frequency=22.0,
            phase=856.6789,
            gains=(3.0, -1.0, 5.0, 1.0),
            sine_generator_index=8,
        )



import pytest
from itertools import cycle

from zhinst.toolkit.driver.modules.impedance_module import (
    ImpedanceModule,
    CalibrationStatus,
)


@pytest.fixture()
def impedance_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_impedance_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.impedanceModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    yield ImpedanceModule(mock_connection.return_value.impedanceModule(), session)


def test_repr(impedance_module):
    assert "MagicMock(DataServerSession(localhost:8004))" in repr(impedance_module)


def test_wait_done(impedance_module):
    calibrate = 1
    status = cycle([1])
    expectedstatus = 1

    def getInt(node, *args, **kwargs):
        if node == "/calibrate":
            return calibrate
        if node == "/status":
            return next(status)
        if node == "/expectedstatus":
            return expectedstatus
        raise RuntimeError("Node not found")

    impedance_module.raw_module.getInt.side_effect = getInt
    impedance_module.raw_module.progress.return_value = [1]
    impedance_module.wait_done()

    # waiting a moment
    status = cycle([0, 0, 0, 1, 1, 1])
    impedance_module.wait_done(sleep_time=0.001)


def test_wait_done_timeout(impedance_module):
    impedance_module.raw_module.getInt.return_value = 0
    impedance_module.raw_module.progress.return_value = [0]
    with pytest.raises(TimeoutError):
        impedance_module.wait_done()


def test_wait_done_fail(impedance_module):
    calibrate = 0
    status = 2
    expectedstatus = 1

    def getInt(node, *args, **kwargs):
        if node == "/calibrate":
            return calibrate
        if node == "/status":
            return status
        if node == "/expectedstatus":
            return expectedstatus
        raise RuntimeError("Node not found")

    impedance_module.raw_module.getInt.side_effect = getInt
    impedance_module.raw_module.progress.return_value = [1]
    with pytest.raises(RuntimeError):
        impedance_module.wait_done()


def test_wait_done_fail_stage(impedance_module):
    calibrate = 0
    status = 2
    expectedstatus = 1

    def getInt(node, *args, **kwargs):
        if node == "/calibrate":
            return calibrate
        if node == "/status":
            return status
        if node == "/expectedstatus":
            return expectedstatus
        raise RuntimeError("Node not found")

    impedance_module.raw_module.getInt.side_effect = getInt
    impedance_module.raw_module.progress.return_value = [1]
    with pytest.raises(RuntimeError):
        impedance_module.wait_done(4)


def test_finish(impedance_module):
    impedance_module.finish()
    impedance_module.raw_module.finish.assert_called_once()


def test_finished(impedance_module):
    impedance_module.raw_module.getInt.side_effect = [1, 1]
    assert impedance_module.finished()

    impedance_module.raw_module.getInt.side_effect = [3, 3]
    assert impedance_module.finished()

    impedance_module.raw_module.getInt.side_effect = [3, 1]
    assert not impedance_module.finished()


def test_finished_step(impedance_module):
    impedance_module.raw_module.getInt.return_value = 1
    assert impedance_module.finished(0)

    assert not impedance_module.finished(1)
    impedance_module.raw_module.getInt.return_value = 3
    assert impedance_module.finished(1)


def test_progress(impedance_module):
    impedance_module.progress()
    impedance_module.raw_module.progress.assert_called_once()


def test_calibrate_status_repr():
    assert (
        repr(CalibrationStatus(13))
        == "step 0: True, step 1: False, step 2: True, step 3: True"
    )
    assert repr(CalibrationStatus(0)) == ""
    assert repr(CalibrationStatus(5)) == "step 0: True, step 1: False, step 2: True"


def test_calibrate_status_item():
    status = CalibrationStatus(13)
    assert status[0] == 1
    assert status[1] == 0
    assert status[2] == 1
    assert status[3] == 1
    assert status[4] == 0
    assert status[5] == 0


def test_calibrate_status_len():
    assert len(CalibrationStatus(13)) == 4
    assert len(CalibrationStatus(0)) == 0
    assert len(CalibrationStatus(2)) == 2
    assert len(CalibrationStatus(5)) == 3



"""HDAWG Instrument Driver."""
import typing as t

from zhinst.toolkit.driver.devices.base import BaseInstrument
from zhinst.toolkit.driver.nodes.awg import AWG
from zhinst.toolkit.nodetree.helper import (
    create_or_append_set_transaction,
    lazy_property,
)
from zhinst.toolkit.nodetree.node import NodeList
from zhinst.toolkit.exceptions import ToolkitError


class HDAWG(BaseInstrument):
    """High-level driver for the Zurich Instruments HDAWG."""

    def enable_qccs_mode(self, gen: int = 1) -> None:
        """Configure the instrument to work with PQSC.

        This method sets the reference clock source to
        connect the instrument to the PQSC.

        Args:
            gen: The QCCS generation that is being configured.
                 Use 1 for a gen1 system, when only HDAWG and UHFQA are used.
                 In this case, the sample rate is set to 2.4 GSa/s and the DIO
                 interface is configured for connection with the UHFQA.
                 Use 2 for a gen2 system, when only HDAWG and SHFs are used.
                 In this case, the sample rate is set to 2.0 GSa/s.
                 (default: 1)

            .. versionchanged:: 0.6.2: Added gen parameter.

        Raises:
            ToolkitError: If the gen argument is not correct.

        Info:
            Use ``factory_reset`` to reset the changes if necessary
        """
        with create_or_append_set_transaction(self._root):
            # Enable QCCS mode, so the instrument is able
            # to operate with ZSync
            self.dios[0].mode("qccs")

            # Gen1 specific elements
            if gen == 1:
                # Set the sample clock to 2.4 GSa/s
                self.system.clocks.sampleclock.freq(2.4e9)
                # Configure DIO
                # Set interface standard to use on the 32-bit DIO to LVCMOS
                self.dios[0].interface(0)
                # Drive the two most significant bytes of the DIO port
                # so the UHFQA receives triggers from PQSC
                self.dios[0].drive(0b1100)

            # Gen2 specific elements
            elif gen == 2:
                # Set the sample clock to 2.0 GSa/s
                self.system.clocks.sampleclock.freq(2.0e9)
                # Turn off DIO
                self.dios[0].drive(0b0000)

            else:
                raise ToolkitError("Only gen1 or gen2 are supported!")

            # Set ZSync clock to be used as reference
            self.system.clocks.referenceclock.source("zsync")

            # Disable DIO triggering on the AWGs,
            # since it's not needed for ZSync messages
            self.awgs["*"].dio.strobe.slope("off")
            self.awgs["*"].dio.valid.polarity("none")

    @lazy_property
    def awgs(self) -> t.Sequence[AWG]:
        """A Sequence of AWG Cores."""
        return NodeList(
            [
                AWG(
                    self.root,
                    self._tree + ("awgs", str(i)),
                    self.serial,
                    i,
                    self.device_type,
                    self.device_options,
                )
                for i in range(len(self["awgs"]))
            ],
            self._root,
            self._tree + ("awgs",),
        )
